(*********************************
 * target machine definitions
 *********************************)

Require Import EqNat.
Require  Export natutil.
Require Export map.
Require Import Arith.

(** register *)
Inductive reg : Set :=
  (* return value register and return address register *)
  | rv : reg | ra : reg  
  (* argument registers *)
  | ar0 : reg | ar1 : reg | ar2 : reg | ar3 : reg 
  (* temporary registers or caller-save registers *)
  | tr0 : reg | tr1 : reg | tr2 : reg | tr3 : reg | tr4 : reg | tr5 : reg
  (* callee-saved registers *)
  | sr0 : reg | sr1 : reg | sr2 : reg | sr3 : reg | sr4 : reg | sr5 : reg | sr6 : reg | sr7 : reg.

Definition beq_reg (r1 r2 : reg) :=
  match r1, r2 with
    | rv, rv => true
    | ra, ra => true
    | ar0, ar0 => true
    | ar1, ar1 => true
    | ar2, ar2 => true
    | ar3, ar3 => true
    | tr0, tr0 => true
    | tr1, tr1 => true
    | tr2, tr2 => true
    | tr3, tr3 => true
    | tr4, tr4 => true
    | tr5, tr5 => true
    | sr0, sr0 => true
    | sr1, sr1 => true
    | sr2, sr2 => true
    | sr3, sr3 => true
    | sr4, sr4 => true
    | sr5, sr5 => true
    | sr6, sr6 => true
    | sr7, sr7 => true
    | _, _ => false
  end.

(** memory addresses *)
Notation addr := nat (only parsing).

(** resource *)
Inductive res : Set :=
  | regR : reg -> res
  | addrR : addr -> res.

Definition beq_res (rs1 rs2 : res) :=
  match rs1, rs2 with
    | regR r1, regR r2 => beq_reg r1 r2
    | addrR a1, addrR a2 => beq_nat a1 a2
    | _, _ => false
  end.

Lemma eq_res_dec : 
  forall rs1 rs2 : res, {rs1 = rs2} + {rs1 <> rs2}.
Admitted.

(** values *)
Notation val := nat (only parsing).

(** state *)
Module StateSpec <: MapSpec.
  Definition domain := res.
  Definition image := val.
  Definition beq_Domain := eq_res_dec.
End StateSpec.

Module StateMod := MapFun(StateSpec).

Definition state := StateMod.Map.

(** code labels *)
Notation lab := nat (only parsing).

(** instruction ( added later )*)
Inductive sins : Set :=
(*  | addi : reg -> reg -> nat -> sins*)
  | movi : reg -> nat -> sins
  | call : lab -> sins
  | cr : reg -> sins
(*  | ld : reg ->  nat -> reg -> sins
  | st : nat -> reg -> reg -> sins
  | alloc : reg -> nat -> sins
  | free : reg -> nat -> sins*).

Inductive bins : Set :=
  | bgtz : reg -> lab -> bins.

Inductive jins : Set :=
  | jmp : lab -> jins
  | ret : jins.

Inductive ins : Set :=
  | sinsI : sins -> ins
  | binsI : bins -> ins
  | jinsI : jins -> ins.

(** code heap *)
Module CodeSpec <: MapSpec.
  Definition domain := lab.
  Definition image := ins.
  Definition beq_Domain := eq_nat_dec.
End CodeSpec.

Module CodeMod := MapFun(CodeSpec).

Definition code := CodeMod.Map.

(** program counter *)
Definition pcounter := lab.

(** index *)
Definition index := nat.

(** world *)
Definition world := (code * state * index * pcounter)%type.

Definition w_cd (w : world) :=
  match w with 
    (c,_,_,_) => c
  end.

Definition w_st (w : world) :=
  match w with
    (_,s,_,_) => s
  end.

Definition w_ind (w : world) :=
  match w with
    (_,_,i,_) => i
  end.

Definition w_pc (w : world) :=
  match w with
    (_,_,_,pc) => pc
  end.

Fixpoint swriteN (n : nat) (s : state) (a : addr) (l : addr -> val) {struct n} : state :=
  match n with 
    | O => s
      | S m => swriteN m (StateMod.put s (addrR a) (l 0)) (S a) (fun x : addr => l (S x))
  end.

Fixpoint sfreeN (n : nat) (s : state) (a : addr) {struct n} : state :=
  match n with 
    | O => s
      | S m => sfreeN m (StateMod.remove s (addrR a)) (S a)
  end.

Inductive condS : sins -> state -> Prop :=
  | condS_movi : 
    forall (s : state) (r : reg) (n : nat),
    StateMod.in_dom (regR r) s
    -> condS (movi r n) s
.

Inductive nextS : sins -> state -> state -> Prop :=
(*  | nextS_addi :
    forall (s : state) (r1 r2 : reg) (n : nat) (v : val),
      s (regR r1) = Some v 
      -> StateMod.in_dom (regR r1) s
      -> nextS (addi r1 r2 n) s (StateMod.put s (regR r1) (v+n))
*)
  | nextS_movi :
    forall (s : state) (r : reg) (n : nat),
      nextS (movi r n) s (StateMod.put s (regR r) n)
(*  | nextS_ld :
    forall (s : state) (r1 r2 : reg) (n : nat) (v1 v2 : val),
      s (regR r2) = Some v1
      -> s (addrR (v1+n)) = Some v2
      -> StateMod.in_dom (regR r1) s
      -> nextS (ld r1 n r2) s (StateMod.put s (regR r1) v2)
  | nextS_st :
    forall (s : state) (r1 r2 : reg) (n : nat) (v1 v2 : val),
      s (regR r1) = Some v1
      -> s (regR r2) = Some v2
      -> StateMod.in_dom (addrR (v1+n)) s
      -> nextS (st n r1 r2) s (StateMod.put s (addrR (v1+n)) v2)
  | nextS_alloc :
    forall (s : state) (r : reg) (n : nat) (a : addr) (l : addr -> val),
      StateMod.in_dom (regR r) s
      -> (forall m : nat, m < n -> StateMod.not_in_dom (addrR (a+m)) s)
      -> nextS (alloc r n) s (StateMod.put (swriteN n s a l) (regR r) a)
  | nextS_free :
    forall (s: state) (r : reg) (n : nat) (a : addr),
      s (regR r) = Some a
      -> (forall m : nat, m < n -> StateMod.in_dom (addrR (a+m)) s)
      -> nextS (alloc r n) s (sfreeN n s a)
*).

Inductive step : world -> world -> Prop :=
  | step_sins : 
    forall (c : code) (s s' : state) (sc : sins) (i : index) (pc : pcounter),
      c pc = Some (sinsI sc) 
      -> nextS sc s s' 
      -> step (c, s, S i, pc) (c, s', i, S pc)
  | step_call :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter),
      c pc = Some (sinsI (call l)) 
      -> StateMod.in_dom (regR ra) s
      -> step (c, s, S i, pc) (c, StateMod.put s (regR ra) (S pc), i, l)
  | step_cr : 
    forall (c : code) (s : state) (r : reg) (i : index) (pc : pcounter) (l : lab),
      c pc = Some (sinsI (cr r))
      -> s (regR r) = Some l
      -> StateMod.in_dom (regR ra) s
      -> step (c, s, S i, pc) (c, StateMod.put s (regR ra) (S pc), i, l)
  | step_bgtz_t : 
    forall (c : code) (s : state) (r : reg) (l : lab) (i : index) (pc : pcounter) (v : val),
      c pc = Some (binsI (bgtz r l))
      -> s ( regR r) = Some v
      -> v > 0
      -> step (c, s, S i, pc) (c, s, i, l)
  | step_bgtz_f : 
    forall (c : code) (s : state) (r : reg) (l : lab) (i : index) (pc : pcounter) (v : val),
      c pc = Some (binsI (bgtz r l))
      -> s ( regR r) = Some v
      -> ~ (v > 0)
      -> step (c, s, S i, pc) (c, s, i, (S pc))
  | step_jmp :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter),
      c pc = Some (jinsI (jmp l))
      -> step (c, s, S i, pc) (c, s, i, l)
  | step_ret :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter),
      c pc = Some (jinsI ret)
      -> s (regR ra) = Some l
      -> step (c, s, S i, pc) (c, s, i, l).

(* failed *)
(*Fixpoint ffinish (c : code) (s s' : state) (i j : index) (pc pc' : pcounter) {struct i} : Prop :=
  exists cm : ins, CodeMod.maps_to c pc cm /\
  match (i, cm) with
    | (O, _) => False
    | (S i', sinsI (call l)) => 
      exists s'' : state, exists j' : index, exists pc'' : pcounter,
      j' < i' /\StateMod.in_dom (regR ra) s /\ ffinish c (StateMod.put s (regR ra) (S pc)) s'' i' j' l pc'' /\
      ffinish c s'' s' j' j pc'' pc'
    | (S i', sinsI (cr r)) =>  
      exists l : lab, exists s'' : state, exists j' : index, exists pc'' : pcounter,
      j' < i' /\ StateMod.maps_to s (regR r) l /\ StateMod.in_dom (regR ra) s /\ 
      ffinish c (StateMod.put s (regR ra) (S pc)) s'' i' j' l pc'' /\ ffinish c s'' s' j' j pc'' pc'
    | (S i', sinsI sc) => 
      exists s'' : state, nextS sc s s'' /\ ffinish c s'' s' i' j (S pc) pc'
    | (S i', binsI (bgtz r l)) =>
      exists v, StateMod.maps_to s (regR r) v /\
        (v > 0 /\ ffinish c s s' i' j l pc' \/ ~(v > 0) /\ ffinish c s s' i' j (S pc) pc')
    | (S i', jinsI (jmp l)) => ffinish c s s' i' j l pc'
    | (S i', jinsI ret) =>
      exists l : lab, StateMod.maps_to s (regR ra) l /\ ffinish c s s' i' j l pc'
  end.
Fixpoint ifinish (c : code) (s s': state) (i j : index) (pc pc' : pcounter) {struct i} : Prop :=
  exists sc : sins, CodeMod.maps_to c pc (sinsI sc) /\
  match (i, sc) with 
    | (O, _) => False
    | (S i', call l) => 
      StateMod.in_dom (regR ra) s /\  ffinish c (StateMod.put s (regR ra) (S pc)) s' i' j l pc'
    | (S i', cr r) =>  
      exists l : lab, StateMod.maps_to s (regR r) l /\ StateMod.in_dom (regR ra) s /\
        ffinish c (StateMod.put s (regR ra) (S pc)) s' i' j l pc'
    | (S i', _) => 
      nextS sc s s' /\ j = i' /\ pc' = S pc
  end
*) 
(** procedure semantics *)

(** instruction sequence *)
Inductive iseq : Set :=
  | sinsIS : sins -> iseq -> iseq
  | binsIS : bins -> iseq -> iseq
  | jinsIS : jins -> iseq.

(** C[l]= I *)
Inductive lookupC : code -> lab -> iseq -> Prop :=
  | lookupC_s :
    forall (c : code) (l : lab) (sc : sins) (is : iseq),
      CodeMod.maps_to c l (sinsI sc) -> lookupC c (S l) is -> lookupC c l (sinsIS sc is)
  | lookupC_b :
    forall (c : code) (l : lab) (bc : bins) (is : iseq),
      CodeMod.maps_to c l (binsI bc) -> lookupC c (S l) is -> lookupC c l (binsIS bc is)
  | lookupC_j :
    forall (c : code) (l : lab) (jc : jins),
      CodeMod.maps_to c l (jinsI jc) -> lookupC c l (jinsIS jc).

Lemma lookupC_same :
  forall (c : code) (l : lab) (is is' : iseq),
    lookupC c l is -> lookupC c l is' -> is = is'.
Proof.
intros.
generalize is' H0; clear is' H0.
induction H; intros.
inversion_clear H1.
assert (I : is = is0).
apply IHlookupC; assumption.
unfold CodeMod.maps_to in *.
rewrite H in H2; inversion H2.
subst; reflexivity.
unfold CodeMod.maps_to in *.
rewrite H in H2; inversion H2.
unfold CodeMod.maps_to in *.
rewrite H in H2; inversion H2.
inversion_clear H1.
unfold CodeMod.maps_to in *.
rewrite H in H2; inversion H2.
assert (I : is = is0).
apply IHlookupC; assumption.
unfold CodeMod.maps_to in *.
rewrite H in H2; inversion H2.
subst; reflexivity.
unfold CodeMod.maps_to in *.
rewrite H in H2; inversion H2.
inversion_clear H0.
unfold CodeMod.maps_to in *.
rewrite H in H1; inversion H1.
unfold CodeMod.maps_to in *.
rewrite H in H1; inversion H1.
unfold CodeMod.maps_to in *.
rewrite H in H1; inversion_clear H1; reflexivity.
Qed.

(** Program *)
Inductive prog : Set :=
  | sinsP : sins -> prog
  | iseqP : iseq -> prog.

Inductive finish : prog -> world -> world -> Prop :=
  | fin_sins : 
    forall (c : code) (s s' : state) (sc : sins) (i : index) (pc : pcounter),
      nextS sc s s' 
      -> finish (sinsP sc) (c, s, S i, pc) (c, s', i, S pc)
  | fin_call :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter) (is : iseq) (w : world),
      lookupC c l is
      -> finish (iseqP is) (c, StateMod.put s (regR ra) (S pc), i, l) w
      -> finish (sinsP (call l)) (c, s, S i, pc) w
  | fin_cr : 
    forall (c : code) (s : state) (r : reg) (i : index) (pc : pcounter) (l : lab) (is : iseq) (w : world),
      s (regR r) = Some l
      -> lookupC c l is
      -> finish (iseqP is) (c, StateMod.put s (regR ra) (S pc), i, l) w
      -> finish (sinsP (cr r)) (c, s, S i, pc) w
  | fin_seq : 
    forall (c : code) (s : state) (sc : sins) (i : index) (pc : pcounter) (is : iseq) (w w' : world),
      finish (sinsP sc) (c, s, i, pc) w
      -> finish (iseqP is) w w'
      -> finish (iseqP (sinsIS sc is)) (c, s, i, pc) w'
  | fin_bgtz_t : 
    forall (c : code) (s : state) (r : reg) (l : lab) (i : index) (pc : pcounter) (v : val) (is is' :iseq) (w : world),
      s ( regR r) = Some v
      -> lookupC c l is'
      -> v > 0
      -> finish (iseqP is') (c, s, i, l) w
      -> finish (iseqP (binsIS (bgtz r l) is))(c, s, S i, pc) w
  | fin_bgtz_f : 
    forall (c : code) (s : state) (r : reg) (l : lab) (i : index) (pc : pcounter) (v : val) (is : iseq) (w : world),
      s ( regR r) = Some v
      -> ~ (v > 0)
      -> finish (iseqP is) (c, s, i, (S pc)) w
      -> finish (iseqP (binsIS (bgtz r l) is)) (c, s, S i, pc) w
  | fin_jmp :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter) (is : iseq) (w : world),
      lookupC c l is
      -> finish (iseqP is) (c, s, i, l) w
      -> finish (iseqP (jinsIS (jmp l)))(c, s, S i, pc) w 
  | fin_ret :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter),
      s (regR ra) = Some l
      -> finish (iseqP (jinsIS ret)) (c, s, S i, pc) (c, s, i, l).

Lemma fin_index_less :
  forall (t :prog) (w w' :world), finish t w w' -> w_ind w > w_ind w'.
Proof.
intros.
unfold w_ind.
induction H.
apply gt_Sn_n.
apply gt_trans with i; [apply gt_Sn_n | exact IHfinish].
apply gt_trans with i; [apply gt_Sn_n | exact IHfinish].
apply gt_trans with (w_ind w); assumption.
apply gt_trans with i; [apply gt_Sn_n | exact IHfinish].
apply gt_trans with i; [apply gt_Sn_n | exact IHfinish].
apply gt_trans with i; [apply gt_Sn_n | exact IHfinish].
apply gt_Sn_n.
Qed.

Lemma ret_pc :
  forall (is : iseq) (w w' : world),
    finish (iseqP is) w w' -> (w_st w') (regR ra) = Some (w_pc w').
Proof.
unfold w_st, w_pc.
intros.
assert (I : exists t : prog, t = iseqP is).
exists (iseqP is); reflexivity.
destruct I as [t I].
rewrite <- I in H.
generalize is I; clear is I.
induction H; intros.
inversion I.
inversion I.
inversion I.
apply IHfinish2 with is; reflexivity.
apply IHfinish with is'; reflexivity.
apply IHfinish with is; reflexivity.
apply IHfinish with is; reflexivity.
assumption.
Qed.

Lemma fin_index :
  forall (t : prog) (c c' : code) (s s' : state) (i j j' : index) (pc pc' : pcounter),
    finish t (c, s, j, pc) (c', s', j', pc')
    -> i > j
    -> finish t (c, s, i, pc) (c', s', j'+i-j, pc').
Proof.
intros.
induction i.
inversion H0.
assert (I : exists w : world, w = (c, s, j, pc)).
exists (c, s, j, pc); reflexivity.
assert (J : exists w' : world, w' = (c', s', j', pc')).
exists (c', s', j', pc'); reflexivity.
destruct I as [w I]; destruct J as [w' J].
rewrite <- I in H; rewrite <- J in H.
generalize c c' s s' i j j' pc pc' H0 I J;
clear c c' s s' i j j' pc pc' H0 I J IHi.
induction H; intros.
inversion I; subst.
inversion J; subst.
replace (j' + S i0 - S j') with i0.
apply fin_sins; assumption.
omega.
inversion I; subst.
apply fin_call with is.
assumption.
induction i0.
assert (I0 : 0 > i).
apply gt_S_n; assumption.
inversion I0.
replace (j' + S (S i0) - S i) with (j' + S i0 -i).
apply IHfinish.
apply gt_S_n; assumption.
reflexivity.
reflexivity.
omega.
inversion I; subst.
apply fin_cr with l is.
assumption.
assumption.
induction i0.
assert (I0 : 0 > i).
apply gt_S_n; assumption.
inversion I0.
replace (j' + S (S i0) - S i) with (j' + S i0 -i).
apply IHfinish.
apply gt_S_n; assumption.
reflexivity.
reflexivity.
omega.
inversion I; subst.
apply fin_seq with (w_cd w, w_st w, w_ind w + S i0 - j, w_pc w).
apply IHfinish1.
assumption.
reflexivity.
destruct w as [[[cx sx] ix] pcx].
reflexivity.
replace (j' + S i0 - j) with (j' + (w_ind w + S i0 -j) - (w_ind w)).
replace (w_ind w + S i0 - j) with (S (w_ind w + i0 - j)).
apply IHfinish2.
omega.
destruct w as [[[cx sx] ix] pcx].
reflexivity.
reflexivity.
omega.
omega.
inversion I; subst.
apply fin_bgtz_t with v is'.
assumption.
assumption.
assumption.
induction i0.
assert (I0 : 0 > i).
apply gt_S_n; assumption.
inversion I0.
clear IHi0.
replace (j' + S (S i0) - S i) with (j' + S i0 - i).
apply IHfinish.
apply gt_S_n; assumption.
reflexivity.
reflexivity.
omega.
inversion I; subst.
apply fin_bgtz_f with v.
assumption.
assumption.
induction i0.
assert (I0 : 0 > i).
apply gt_S_n; assumption.
inversion I0.
clear IHi0.
replace (j' + S (S i0) - S i) with (j' + S i0 - i).
apply IHfinish.
apply gt_S_n; assumption.
reflexivity.
reflexivity.
omega.
inversion I; subst.
apply fin_jmp with is.
assumption.
induction i0.
assert (I0 : 0 > i).
apply gt_S_n; assumption.
inversion I0.
clear IHi0.
replace (j' + S (S i0) - S i) with (j' + S i0 - i).
apply IHfinish.
apply gt_S_n; assumption.
reflexivity.
reflexivity.
omega.
inversion I; subst.
inversion J; subst.
replace (j' + S i0 - S j') with i0.
apply fin_ret.
assumption.
omega.
Qed.

Lemma fin_c : 
  forall (t : prog) (w w' : world),
    finish t w w' -> w_cd w = w_cd w'.
intros.
induction H.
reflexivity.
exact IHfinish.
exact IHfinish.
rewrite <- IHfinish1 in IHfinish2; exact IHfinish2. 
exact IHfinish.
exact IHfinish.
exact IHfinish.
reflexivity.
Qed.

(** safe to execut program *)
Inductive safe : prog -> world -> Prop :=
  | safe_zero : 
    forall (t : prog) (c : code) (s : state) (pc : pcounter),
      safe t (c, s, 0, pc)
  | safe_sins : 
    forall (c : code) (s : state) (sc : sins) (i : index) (pc : pcounter),
      c pc = Some (sinsI sc) 
      -> condS sc s 
      -> safe (sinsP sc) (c, s, S i, pc)
  | safe_call :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter) (is : iseq),
      c pc = Some (sinsI (call l)) 
      -> StateMod.in_dom (regR ra) s
      -> lookupC c l is
      -> safe (iseqP is) (c, StateMod.put s (regR ra) (S pc), i, l)
      -> (forall w' : world, finish (iseqP is) (c, StateMod.put s (regR ra) (S pc), i, l) w' -> w_pc w' = S pc)
      -> safe (sinsP (call l)) (c, s, S i, pc)
  | safe_cr : 
    forall (c : code) (s : state) (r : reg) (i : index) (pc : pcounter) (l : lab) (is :iseq),
      c pc = Some (sinsI (cr r))
      -> s (regR r) = Some l
      -> lookupC c l is
      -> StateMod.in_dom (regR ra) s
      -> safe (iseqP is) (c, StateMod.put s (regR ra) (S pc), i, l)
      -> (forall w' : world, finish (iseqP is) (c, StateMod.put s (regR ra) (S pc), i, l) w' -> w_pc w' = S pc)
      -> safe (sinsP (cr r)) (c, s, S i, pc)
  | safe_seq : 
    forall (c : code) (s : state) (sc : sins) (is : iseq) (i : index) (pc : pcounter),
      lookupC c pc (sinsIS sc is)
      -> safe (sinsP sc) (c, s, i, pc)
      -> (forall w : world, finish (sinsP sc) (c, s, i, pc) w -> safe (iseqP is) w)
      -> safe (iseqP (sinsIS sc is))(c, s, i, pc)
  | safe_bgtz : 
    forall (c : code) (s : state) (r : reg) (l : lab) (is is': iseq) (i : index) (pc : pcounter) (v : val),
      lookupC c pc (binsIS (bgtz r l) is)
      -> lookupC c l is'
      -> s ( regR r) = Some v
      -> (v > 0 -> safe (iseqP is') (c, s, i, l)) 
      -> (~(v > 0) -> safe (iseqP is) (c, s, i, S pc))
      -> safe (iseqP (binsIS (bgtz r l) is)) (c, s, S i, pc)
  | safe_jmp :
    forall (c : code) (s : state) (l : lab) (is : iseq) (i : index) (pc : pcounter),
      lookupC c pc (jinsIS (jmp l))
      -> lookupC c l is
      -> safe (iseqP is) (c, s, i, l) 
      -> safe (iseqP (jinsIS (jmp l))) (c, s, S i, pc) 
  | safe_ret :
    forall (c : code) (s : state) (i : index) (pc : pcounter),
      lookupC c pc (jinsIS ret)
      -> StateMod.in_dom (regR ra) s
      -> safe (iseqP (jinsIS ret)) (c, s, S i, pc).

(**
Inductive ifinish : world -> world -> Prop :=
  | ifin_sins : 
    forall (c : code) (s s' : state) (sc : sins) (i : index) (pc : pcounter),
      c pc = Some (sinsI sc) 
      -> nextS sc s s' 
      -> ifinish (c, s, S i, pc) (c, s', i, S pc)
  | ifin_call :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter) (w : world),
      c pc = Some (sinsI (call l)) 
      -> StateMod.in_dom (regR ra) s
      -> ffinish (c, StateMod.put s (regR ra) (S pc), i, l) w
      -> ifinish (c, s, S i, pc) w
  | ifin_cr : 
    forall (c : code) (s : state) (r : reg) (i : index) (pc : pcounter) (l : lab) (w : world),
      c pc = Some (sinsI (cr r))
      -> s (regR r) = Some l
      -> StateMod.in_dom (regR ra) s
      -> ffinish (c, StateMod.put s (regR ra) (S pc), i, l) w
      -> ifinish (c, s, S i, pc) w
with ffinish : world -> world -> Prop :=
  | ffin_sins : 
    forall (c : code) (s : state) (sc : sins) (i : index) (pc : pcounter) (w w' : world),
      c pc = Some (sinsI sc)
      -> ifinish (c, s, i, pc) w
      -> ffinish w w'
      -> ffinish (c, s, i, pc) w'
  | ffin_bgtz_t : 
    forall (c : code) (s : state) (r : reg) (l : lab) (i : index) (pc : pcounter) (v : val) (w : world),
      c pc = Some (binsI (bgtz r l))
      -> s ( regR r) = Some v
      -> v > 0
      -> ffinish (c, s, i, l) w
      -> ffinish (c, s, S i, pc) w
  | ffin_bgtz_f : 
    forall (c : code) (s : state) (r : reg) (l : lab) (i : index) (pc : pcounter) (v : val) (w : world),
      c pc = Some (binsI (bgtz r l))
      -> s ( regR r) = Some v
      -> ~ (v > 0)
      -> ffinish (c, s, i, (S pc)) w
      -> ffinish (c, s, S i, pc) w
  | ffin_jmp :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter) (w : world),
      c pc = Some (jinsI (jmp l))
      -> ffinish (c, s, i, l) w
      -> ffinish (c, s, S i, pc) w 
  | ffin_ret :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter),
      c pc = Some (jinsI ret)
      -> s (regR ra) = Some l
      -> ffinish (c, s, S i, pc) (c, s, i, l).

Lemma ffin_index_less :
  forall (w w' :world), ffinish w w' -> w_ind w > w_ind w'.
Admitted.

Lemma ifin_index_less :
  forall (w w' : world),
    ifinish w w' -> w_ind w > w_ind w'.
Admitted.



(** safe to execute a sequencial instruction *)
Inductive isafe : world -> Prop :=
  | isafe_zero : 
    forall (c : code) (s : state) (pc : pcounter),
      isafe  (c, s, O, pc)
  | isafe_sins : 
    forall (c : code) (s s' : state) (sc : sins) (i : index) (pc : pcounter),
      c pc = Some (sinsI sc) 
      -> nextS sc s s' 
      -> isafe (c, s, S i, pc)
  | isafe_call :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter),
      c pc = Some (sinsI (call l)) 
      -> StateMod.in_dom (regR ra) s
      -> fsafe (c, StateMod.put s (regR ra) (S pc), i, l)
      -> (forall w' : world, ffinish (c, StateMod.put s (regR ra) (S pc), i, l) w' -> w_pc w' = S pc)
      -> isafe (c, s, S i, pc)
  | isafe_cr : 
    forall (c : code) (s : state) (r : reg) (i : index) (pc : pcounter) (l : lab),
      c pc = Some (sinsI (cr r))
      -> s (regR r) = Some l
      -> StateMod.in_dom (regR ra) s
      -> fsafe (c, StateMod.put s (regR ra) (S pc), i, l)
      -> (forall w' : world, ffinish (c, StateMod.put s (regR ra) (S pc), i, l) w' -> w_pc w' = S pc)
      -> isafe (c, s, S i, pc)
with fsafe : world -> Prop :=
  | fsafe_zero : 
    forall (c : code) (s : state) (pc : pcounter),
      fsafe (c, s, O, pc) 
  | fsafe_sins : 
    forall (c : code) (s : state) (sc : sins) (i : index) (pc : pcounter),
      c pc = Some (sinsI sc)
      -> (forall w : world, ifinish (c, s, i, pc) w -> fsafe w)
      -> fsafe (c, s, i, pc)
  | fsafe_bgtz_t : 
    forall (c : code) (s : state) (r : reg) (l : lab) (i : index) (pc : pcounter) (v : val),
      c pc = Some (binsI (bgtz r l))
      -> s ( regR r) = Some v
      -> (v > 0 -> fsafe (c, s, i, l)) 
      -> (~(v > 0) -> fsafe (c, s, i, S pc))
      -> fsafe (c, s, S i, pc)
  | fsafe_jmp :
    forall (c : code) (s : state) (l : lab) (i : index) (pc : pcounter),
      c pc = Some (jinsI (jmp l))
      -> fsafe (c, s, i, l) 
      -> fsafe (c, s, S i, pc) 
  | fsafe_ret :
    forall (c : code) (s : state) (i : index) (pc : pcounter),
      c pc = Some (jinsI ret)
      -> StateMod.in_dom (regR ra) s
      -> fsafe (c, s, S i, pc).
*)

Lemma fin_safe_sc_pc :
  forall (sc : sins) (w w' : world),
    safe (sinsP sc) w -> finish (sinsP sc) w w' -> w_pc w' = S (w_pc w).
Proof.
intros.
inversion H0.
reflexivity.
rewrite <- H4 in H.
rewrite <- H1 in H.
inversion H.
inversion H12.
apply H15.
assert (I : is = is0).
apply lookupC_same with c l; assumption.
subst.
assumption.
subst sc w.
inversion H.
inversion H11.
apply H15.
rewrite H2 in H11.
inversion H11.
subst l0.
assert (I : is = is0).
apply lookupC_same with c l; assumption.
subst is0.
assumption.
Qed.

Lemma safe_mono :
  forall (t : prog) (c : code) (s : state) (i j : index) (pc : pcounter),
    safe t (c, s, i, pc) 
    -> i > j
    -> safe  t (c, s, j, pc).
Proof.
intros.
induction j.
apply safe_zero.
clear IHj.
assert (I : exists w : world, (w = (c, s, i, pc))).
exists (c, s, i, pc); reflexivity.
destruct I as [w I].
rewrite <- I in H.
generalize c s i j pc H0 I; clear c s i j pc H0 I.
induction H; intros.
inversion I; subst.
inversion H0.
inversion I; subst.
apply safe_sins; assumption.
inversion I; subst.
apply safe_call with is.
 assumption.
assumption.
assumption.
induction j.
apply safe_zero.
apply IHsafe with i.
apply gt_S_n; assumption.
reflexivity.
intros.
assert (I1 : w_pc w' = w_pc (w_cd w', w_st w', w_ind w' + i - j, w_pc w')).
reflexivity.
rewrite I1; apply H3.
apply fin_index.
destruct w' as [[[cx sx] ix] pcx].
assumption.
apply gt_S_n; assumption.
inversion I; subst.
inversion H0.
inversion I; subst.
apply safe_cr with l is.
assumption.
assumption.
assumption.
assumption.
induction j.
apply safe_zero.
apply IHsafe with i.
apply gt_S_n; assumption.
reflexivity.
intros.
assert (I1 : w_pc w' = w_pc (w_cd w', w_st w', w_ind w' + i - j, w_pc w')).
reflexivity.
rewrite I1; apply H4.
apply fin_index.
destruct w' as [[[cx sx] ix] pcx].
assumption.
apply gt_S_n; assumption.
inversion I; subst; clear I.
apply safe_seq.
assumption.
apply IHsafe with i0.
assumption.
reflexivity.
intros.
destruct w as [[[c' s'] i'] pc'].
induction i'.
apply safe_zero.
clear IHi'.
apply H2 with (c', s', (S i') + i0 - (S j), pc') (S i' + i0 - (S j)).
apply fin_index.
assumption.
assumption.
omega.
reflexivity.
inversion I; subst; clear I.
apply safe_bgtz with is' v.
assumption.
assumption.
assumption.
intro.
induction j.
apply safe_zero.
clear IHj.
apply H3 with i.
assumption.
apply gt_S_n; assumption.
reflexivity.
intro.
induction j.
apply safe_zero.
clear IHj.
apply H5 with i.
assumption.
apply gt_S_n; assumption.
reflexivity.
inversion I; subst; clear I.
apply safe_jmp with is.
assumption.
assumption.
induction j.
apply safe_zero.
clear IHj.
apply IHsafe with i.
apply gt_S_n; assumption.
reflexivity.
inversion I; subst; clear I.
apply safe_ret; assumption.
Qed.

(** these two axioms assume that 
  the program instructions are local 
*)
Axiom local_frame :
  forall (t : prog) (c c' : code) (s s1 s2 s2' : state) (i i' : index) (pc pc' : pcounter),
    StateMod.disjUnion s s1 s2 
    -> safe t (c, s, i, pc) 
    -> finish t (c, s2, i, pc) (c', s2', i', pc')
    -> (exists s' : state, StateMod.disjUnion s' s1 s2' /\ finish t (c, s, i, pc) (c', s', i', pc')).

Axiom local_mono :
  forall (t :prog) (c : code) (s s' s'': state) (i : index) (pc : pcounter),
    StateMod.disjUnion s s' s'' -> safe t (c, s, i, pc) -> safe t (c, s'', i, pc). 
